<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\TestBundle\Functional;


use Cybersponse\CoreBundle\Provider\Cacheable;
use Cybersponse\FilesBundle\Document\File;
use Cybersponse\MetadataBundle\Entity\ModelMetadata;
use Doctrine\Common\Persistence\ManagerRegistry;
use Doctrine\ODM\MongoDB\DocumentManager;
use Nelmio\Alice\Fixtures;
use Symfony\Bundle\FrameworkBundle\Client;
use Symfony\Component\DependencyInjection\ContainerInterface;

abstract class ApiTestCase extends JwtWebTestCase
{
    /** @var string */
    protected $actorFixtureName = 'mockActorSecurityAdmin';

    /** @var ContainerInterface */
    protected static $container;

    /** @var Client */
    protected $client;

    /** @var array */
    protected $fixtures = [];

    /** @var bool */
    protected $rebuildDatabases = true;

    /** @var string */
    protected $requestVerb = 'GET';

    /** @var string */
    protected $requestUri = '/';

    /** @var array */
    protected $requestParameters = [];

    /** @var array */
    protected $requestFiles = [];

    /** @var null|string */
    protected $requestContent = null;

    /** @var bool */
    protected $clearCustomMetadataCache = false;

    /**
     * @param array $requestFiles
     */
    public function setRequestFiles($requestFiles)
    {
        $this->requestFiles = $requestFiles;
    }

    /**
     * @return string
     */
    protected function getRequestUri()
    {
        return $this->requestUri;
    }

    /**
     * @param array $content
     */
    protected function setJsonRequestContent(array $content)
    {
        $this->requestContent = json_encode($content);
    }

    protected function setUp()
    {
        if ($this->clearCustomMetadataCache) {
            $container = $this->getContainer();
            /** @var \Predis\Client $client */
            $client = $container->get('snc_redis.deletes');
            $items = $client->keys("test:default:\\[Cybersponse\\\\TestBundle*");
            $client->del($items);

            $this->clearCustomMetadataCache = false;
        }

        if ($this->rebuildDatabases) {
            $metadataOdmFixtures = $this->loadOdmFixtureFiles([
                '@CybersponseTestBundle/Resources/fixtures/ODM/metadata.yml'
            ], false, 'metadata');
            $defaultOdmFixtures = $this->loadOdmFixtureFiles([
                '@CybersponseTestBundle/Resources/fixtures/ODM/default.yml'
            ], false, 'default');
            $defaultOrmFixtures = $this->loadFixtureFiles([
                '@CybersponseTestBundle/Resources/fixtures/ORM/default.yml'
            ], false, 'default');
            $this->actor = $defaultOrmFixtures[$this->actorFixtureName];

            $this->fixtures = [
                'ODM' => [
                    'metadata' => $metadataOdmFixtures,
                    'default' => $defaultOdmFixtures
                ],
                'ORM' => [
                    'default' => $defaultOrmFixtures
                ]
            ];

            $this->clearCaches($this->getContainer());
        }

        if (!isset(static::$container)) {
            static::$container = $this->getContainer();
        }

        parent::setUp();
    }

    protected function loadOdmFixtureFiles(array $paths = array(), $append = false, $omName = null)
    {
        if (!class_exists('Nelmio\Alice\Fixtures')) {
            // This class is available during tests, no exception will be thrown.
            // @codeCoverageIgnoreStart
            throw new \BadMethodCallException('nelmio/alice should be installed to use this method.');
            // @codeCoverageIgnoreEnd
        }

        $container = $this->getContainer();

        /** @var ManagerRegistry $registry */
        $registry = $container->get('doctrine_mongodb');

        /** @var DocumentManager $documentManager */
        $documentManager = $registry->getManager($omName);

        if ($append === false) {
            $schemaManager = $documentManager->getSchemaManager();

            if ($omName == 'default') {
                //Files are not dropped properly with the `dropCollections` method
                $fileDatabase = $documentManager->getDocumentDatabase(File::class);
                $fileDatabase->dropCollection('File.files');
                $fileDatabase->dropCollection('File.chunks');
            }

            $schemaManager->dropCollections();
            $schemaManager->createCollections();
        }

        $files = array();

        $kernel = $this->getContainer()->get('kernel');

        foreach ($paths as $path) {
            if ($path[0] !== '@' && file_exists($path) === true) {
                $files[] = $path;
                continue;
            }

            $files[] = $kernel->locateResource($path);
        }

        // Check if the Hautelook AliceBundle is registered and if yes, use it instead of Nelmio Alice
        $hautelookLoaderServiceName = 'hautelook_alice.fixtures.loader';
        if ($container->has($hautelookLoaderServiceName)) {
            $loaderService = $container->get($hautelookLoaderServiceName);
            $persisterClass = 'Nelmio\Alice\Persister\Doctrine';

            return $loaderService->load(new $persisterClass($documentManager), $files);
        }

        return Fixtures::load($files, $documentManager);
    }

    protected function executeSystemUnderTest()
    {
        $this->client = static::makeClient(true);

        $this->client->request(
            $this->requestVerb,
            $this->getRequestUri(),
            $this->requestParameters,
            $this->requestFiles,
            [],
            $this->requestContent
        );
    }

    /**
     * @return null|\Symfony\Component\HttpFoundation\Response
     */
    function getResponse()
    {
        return $this->client->getResponse();
    }

    function getFileResponse()
    {
        return $this->client->getInternalRequest()->getFiles();
    }

    protected function tearDown()
    {
        if (isset($this->client)) {
            $container = $this->client->getContainer();

            $this->clearCaches($container);

            unset($this->client);
        }

        if ($this->rebuildDatabases) {
            unset($this->fixtures);
        }

        parent::tearDown();
    }

    /**
     * @param ContainerInterface $container
     */
    protected function clearCaches($container)
    {
        $cacheServiceNames = [
            'cybersponse.authorization.permissions.permissions_dictionary_cache',
            'cybersponse.authorization.permissions.permissions_list_cache',
            'cybersponse.authorization.ownership.accessible_team_ids_cache',
            'cybersponse.authentication.hmac_key_cache'
        ];
        foreach ($cacheServiceNames as $cacheServiceName) {
            /** @var Cacheable $cache */
            $cache = $container->get($cacheServiceName);
            $cache->clearCache();
        }
    }

    public static function tearDownAfterClass()
    {
        if (isset(static::$container)) {
            /** @var ManagerRegistry $registry */
            $registry = static::$container->get('doctrine_mongodb');

            /** @var DocumentManager $documentManager */
            $documentManager = $registry->getManagerForClass(ModelMetadata::class);

            $schemaManager = $documentManager->getSchemaManager();

            $schemaManager->dropCollections();
        }
    }


}
